import numpy as np
from abc import ABC, abstractmethod
import scipy.special as special
from numba import jit, vectorize


class LineShape(ABC):
    """
    Base class for implementing line shapes.  A line shape represents integration across a high resolution measurement
    down to a lower resolution measurement.
    """

    @abstractmethod
    def integration_weights(self, mean: float, available_samples: np.ndarray, normalize=True):
        """
        Integration weights for the line shape.

        Parameters
        ----------
        mean : float
            Value to integrate to

        available_samples : np.ndarray
            Array of sample values that are available to use in the integration.

        normalize : bool, Optional
            If true, resulting weights are normalized such that np.nansum(weights) = 1

        Returns
        -------
        np.ndarray
            Integration weights, same size as available_samples.
        """
        pass

    def bounds(self):
        """
        Boundaries of the line shape.  Values outside this range are 0

        Returns
        -------
        (left, right)
            Left and right boundaries of the line shape
        """
        pass


class Gaussian(LineShape):
    def __init__(self, fwhm: float = None, stdev: float = None, max_stdev=5, mode='linear'):
        """
        Gaussian line shape.

        Parameters
        ----------
        fwhm : float
            Full width half maximum, only specify one of fwhm or stdev

        stdev : float
            Standard deviation, only specify one of fwhm or stdev

        max_stdev: int, optional
            Values farther than max_stdev*stdev are truncated to 0.  Default=5

        mode : string, one of ['constant', 'linear']
             If constant, then the gaussian is sampled at the integration location.  If linear, then the gaussian
             is integrated with a triangular base function representing linear integration across the sample.  Linear
             is much more accurate at the cost of a small performance hit.  Default 'linear'.
        """

        self._fwhm_to_stdev = 1 / (2 * np.sqrt(2 * np.log(2)))

        if fwhm is not None and stdev is not None:
            raise ValueError('Only one of fwhm or stdev should be specified')

        if fwhm is None and stdev is None:
            raise ValueError('One of fwhm or stdev needs to be specified')

        if fwhm is not None:
            self._stdev = fwhm * self._fwhm_to_stdev
        else:
            self._stdev = stdev

        self.max_stdev = max_stdev

        self._mode = mode

    def integration_weights(self, mean: float, available_samples: np.ndarray, normalize=True):
        """
        The lineshape converts a function at high resolution, H, to one at low resolution, L.

        L(mean) = np.dot(H, integration_weights(mean, available_samples))

        Parameters
        ----------
        mean : float
            value (in the range of available_samples) to integrate to
        available_samples : np.ndarray
            Values that the high resolution function is available at
        normalize : bool, optional
            If true, the returned weights will sum to 1

        Returns
        -------
        np.ndarray
            Weights to use in the high resolution integration
        """
        if self._stdev == 0:
            # Special case, just interpolate to the mean
            # TODO: This is nearest neighbor, switch to interpolate
            weights = np.zeros_like(available_samples)

            weights[np.argmin(np.abs(mean - available_samples))] = 1

            return weights

        difference = mean - available_samples

        if self._mode == 'constant':
            gaussian = self._analytic_constant_weights(mean, available_samples)
        elif self._mode == 'linear':
            gaussian = self._analytic_linear_weights(mean, available_samples)
        else:
            raise ValueError('mode must be one of linear or constant')

        gaussian[np.abs(difference) > self.max_stdev * self._stdev] = 0
        # Sometimes numerical funniness can cause negative values
        # TODO: Check if this should be abs or set to 0
        gaussian = np.abs(gaussian)

        if normalize:
            if np.abs(np.sum(gaussian)) > 0:
                gaussian /= np.sum(gaussian)
            else:
                # Should be all zeros, indicating no contribution to the point which is okay
                # Just to be sure
                gaussian = np.zeros_like(gaussian)

        return gaussian

    def bounds(self):
        """
        If integration_weights is called with mean=0, all values outside the range [lower_bound, upper_bound] are
        guaranteed to be 0.

        Returns
        -------
        [lower_bound, upper_bound]
        """
        return -self.max_stdev * self._stdev, self.max_stdev * self._stdev

    def _analytic_linear_weights(self, mean, available_samples):
        return _gaussian_analytic_linear_weights(self.max_stdev, self._stdev, mean, np.ascontiguousarray(available_samples))

    def _analytic_constant_weights(self, mean, available_samples):
        difference = mean - available_samples

        gaussian = np.exp(-0.5 * (difference / self._stdev) ** 2)

        return gaussian


@vectorize('f8(f8)', nopython=True)
def fasterf1(x):
    """
    Fast error function approximation
    """
    p = 0.47047
    a1 = 0.3480242
    a2 = -0.0958798
    a3 = 0.7487556

    t = 1 / (1 + p * np.abs(x))

    return (1 - (a1 * t + a2 * t ** 2 + a3 * t ** 3) * np.exp(-np.abs(x) ** 2)) * np.sign(x)


@vectorize('f8(f8, f8, f8, f8)', nopython=True)
def _gaussian_analytic_linear_weights_helper(width_left, width_right, offsets, stdev):
    """
    See theory/line_shape_integrals.nb for explanation
    """
    stdev /= np.sqrt(0.5)

    result = 0.5 * stdev * (stdev * (np.exp(-(offsets - width_left) ** 2 / stdev ** 2) / width_left +
                                     np.exp(-(offsets + width_right) ** 2 / stdev ** 2) / width_right -
                                     np.exp(-offsets**2 / stdev**2) * (width_left + width_right) / (width_left * width_right))
                            + 1 / (width_left * width_right) * np.sqrt(np.pi) * (-offsets * (width_right + width_left) * fasterf1(offsets / stdev) +
                                                                                 (offsets - width_left) * width_right * fasterf1((offsets - width_left) / stdev) +
                                                                                 width_left * (offsets + width_right) * fasterf1((offsets + width_right) / stdev)))

    return result


def _gaussian_analytic_linear_weights(max_stdev, stdev, mean, available_samples):
    """
    Calculates the weights by integrating the gaussian form with a triangle function
    """
    weights = np.zeros_like(available_samples)

    # Difference between center of the gaussian and the sample
    offsets = mean - available_samples
    within_stdev = np.abs(offsets) < max_stdev * stdev

    # Interpolation width on the left side of each sample
    widths = np.diff(available_samples)
    width_left = np.abs(np.hstack((np.array([widths[0]]), widths)))

    # Interpolation width on the right side of each sample
    width_right = np.abs(np.hstack((widths, np.array([widths[-1]]))))

    offsets = offsets[within_stdev]
    width_left = width_left[within_stdev]
    width_right = width_right[within_stdev]

    # Formula derived by integrating a triangle function with a gaussian
    weights[within_stdev] = _gaussian_analytic_linear_weights_helper(width_left, width_right, offsets, stdev)

    return weights


class DeltaFunction(LineShape):
    def __init__(self):
        """
        DeltaFunction line shape.  The nearest sample is always taken.

        """
        pass

    def integration_weights(self, mean: float, available_samples: np.ndarray, normalize=True):
        # Interpolate to the mean value
        weights = np.zeros_like(available_samples)

        weights[np.argmin(np.abs(mean - available_samples))] = 1

        return weights

    def bounds(self):
        return 0, 0


class Rectangle(LineShape):
    def __init__(self, width, mode='linear'):
        """
        Rectangular line shape

        Parameters
        ----------
        width : float
            Full width of the line shape.
        """
        self._width = width
        self._mode = mode

    def integration_weights(self, mean: float, available_samples: np.ndarray, normalize=True):
        if self._mode == 'linear':
            weights = self._analytic_linear_weights(mean, available_samples)
        else:
            weights = np.zeros_like(available_samples)

            offsets = mean - available_samples

            weights[np.abs(offsets) < self._width / 2] = 1

        if normalize:
            if np.nansum(weights) == 0:
                raise ValueError()
            weights /= np.nansum(weights)

        return weights

    def _analytic_linear_weights(self, mean, available_samples):
        """
        See theory/line_shape_integrals.nb for more information.  Linear weights are calculated by integrating
        the rectangle function with a triangle function.
        """
        weights = np.zeros_like(available_samples)

        # Difference between center of the gaussian and the sample
        offsets = mean - available_samples
        #within_width = np.abs(offsets) < self._width

        # Interpolation width on the left side of each sample
        widths = np.diff(available_samples)

        width_left = np.abs(np.hstack((np.array([widths[0]]), widths)))

        # Interpolation width on the right side of each sample
        width_right = np.abs(np.hstack((widths, np.array([widths[-1]]))))

        #offsets = offsets[within_width]
        #width_left = width_left[within_width]
        #width_right = width_right[within_width]

        weights += _rectangle_analytic_linear_weights_helper_left(width_left, offsets, self._width)
        weights += _rectangle_analytic_linear_weights_helper_right(width_right, offsets, self._width)

        return weights

    def bounds(self):
        return [-self._width/2, self._width/2]


@vectorize('f8(f8, f8, f8)', nopython=True)
def _rectangle_analytic_linear_weights_helper_left(width_left, offset, rect_width):
    if offset == 0 and rect_width < 2 * width_left:
        return -rect_width * (rect_width - 4*width_left) / (8 * width_left)
    elif (offset > 0 and 2 * offset + rect_width < 2 * width_left and 2 * offset < rect_width) or\
        (offset < 0 and 2 * offset + rect_width > 0 and 2 * offset + rect_width < 2 * width_left):
        return -(2*offset + rect_width) * (2*offset + rect_width - 4 *width_left) / (8*width_left)
    elif offset > 0 and 2 *offset + rect_width <= 2*width_left and (2*offset == rect_width or (rect_width > 0 and 2 *offset >= rect_width)):
        return rect_width - offset*rect_width / width_left
    elif width_left > 0 and ((offset == 0 and rect_width > 2 *width_left) or (2*offset + rect_width >= 2*width_left and offset < 0)or (offset > 0 and 2*offset + rect_width > 2*width_left and 2*offset < rect_width)):
        return width_left / 2
    elif offset > 0 and 2*offset + rect_width > 2 *width_left and ((2*offset == rect_width and width_left < 0) or (2*offset > rect_width and 2*offset < rect_width + 2*width_left)):
        return (-2*offset + rect_width + 2 *width_left)**2 / (8*width_left)
    else:
        return 0


@vectorize('f8(f8, f8, f8)', nopython=True)
def _rectangle_analytic_linear_weights_helper_right(width_right, offset, rect_width):
    if width_right > 0 and ((offset == 0 and rect_width > 0 and rect_width > 2*width_right) or (2*(offset + width_right) < rect_width and ((2*offset + rect_width > 0 and offset < 0) or (offset > 0 and 2 * offset < rect_width)))):
        return width_right / 2
    elif offset < 0 and ((2*offset + rect_width == 0 and 2*(offset + width_right) >= rect_width) or (rect_width > 0 and 2 *(offset + width_right) > rect_width and 2 * offset + rect_width < 0)):
        return rect_width * (offset + width_right) / width_right
    elif offset == 0 and rect_width > 0 and rect_width <= 2 *width_right:
        return -rect_width * (rect_width - 4*width_right) / (8*width_right)
    elif 2*(offset + width_right) >= rect_width and ((2*offset + rect_width > 0 and offset < 0) or (offset > 0 and 2 *offset < rect_width)):
        return -(2*offset - rect_width) * (2*offset - rect_width + 4*width_right) / (8*width_right)
    elif 2*offset + rect_width + 2*width_right > 0 and offset < 0 and ((2*offset + rect_width ==0 and 2*(offset + width_right) < rect_width) or (rect_width > 0 and 2*offset + rect_width < 0 and 2*(offset + width_right) <= rect_width)):
        return (2*offset + rect_width + 2*width_right)**2 / (8*width_right)
    else:
        return 0
