import xarray as xr
from typing import List
from copy import copy
from abc import ABC, abstractmethod
import sasktran as sk
import numpy as np


class RadianceBase(ABC):
    """
    Base functionality that every radiance format must support
    """
    def __init__(self, ds: xr.Dataset):
        self._ds = ds

    @property
    def data(self):
        return self._ds

    @data.setter
    def data(self, value):
        self._ds = value

    def tangent_locations(self):
        """
        Calculates tangent locations for all lines of sight.  If the line of sight does not have a tangent location
        the ground intersection is returned instead
        Returns
        -------
        xr.Dataset
            Dataset containing 'latitude', 'longitude', and 'altitude' of the tangent locations

        """
        los_dims = [dim for dim in self._ds['los_vectors'].dims if dim != 'xyz']

        stacked_los = self._ds['los_vectors'].stack(temp_dim=los_dims)
        stacked_obs = self._ds['observer_position'].stack(temp_dim=los_dims)
        stacked_mjd = self._ds['mjd'].stack(temp_dim=los_dims)

        latitudes = []
        longitudes = []
        altitudes = []
        for idx in stacked_los['temp_dim']:
            one_los = stacked_los.sel(temp_dim=idx)
            one_obs = stacked_obs.sel(temp_dim=idx)
            one_mjd = stacked_mjd.sel(temp_dim=idx)

            los = sk.LineOfSight(mjd=one_mjd.values, observer=one_obs.values, look_vector=one_los.values)

            location = los.tangent_location()

            if location is None:
                location = los.ground_intersection(0)

            latitudes.append(location.latitude)
            longitudes.append(location.longitude)
            altitudes.append(location.altitude)

        result = xr.Dataset({'latitude': (['temp_dim'], latitudes),
                             'longitude': (['temp_dim'], longitudes),
                             'altitude': (['temp_dim'], altitudes)},
                            coords=stacked_los.coords)

        result = result.unstack('temp_dim').drop('xyz')

        return result

    @abstractmethod
    def to_raw(self):
        pass


class RadianceRaw(RadianceBase):
    def __init__(self, ds: xr.Dataset):
        """
        Raw radiance measurements.  This is the simplest structure that can hold measurements from a wide variety of
        instruments.  Each measurement is represented by a line of sight (look vector, observer position, time) and
        a single radiance value.  Optionally, additional parameters may be added such as the estimated noise.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset containing the variables ['radiance', 'los_vector', 'observer_position'] and the coordinates ['meas']
            'meas' is an index that goes across all measurements (could be wavelength or los changes)
        """
        super().__init__(ds)

    def to_raw(self):
        return self

    def _validate_format(self):
        return True


class RadianceGridded(RadianceBase):
    def __init__(self, ds: xr.Dataset):
        """
        A specific radiance format where the radiance data can be represented on a (wavelength, line of sight) grid.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset containing the variable ['radiance', 'los_vector', 'observer_position'] and the coordinates
            ['wavelength', 'los'] where 'los' is a simple indexer that goes across changes in the line of sight.
        """
        super().__init__(ds)

    def _validate_format(self):
        return True

    def to_raw(self):
        new_ds = self.data.stack(meas=['wavelength', 'los'])

        return RadianceRaw(new_ds)
