from skretrieval.retrieval import Minimizer, RetrievalTarget, ForwardModel
from scipy import sparse
import numpy as np
import logging
from copy import copy


class Rodgers(Minimizer):

    def __init__(self, max_iter=10, lm_damping=0, iterative_update_lm=False, retreat_lm=False, lm_change_factor=1.5):
        self._max_iter = max_iter
        self._lm_damping = lm_damping
        self._iterative_update_lm = iterative_update_lm
        self._retreat_lm = retreat_lm
        self._lm_change_factor = lm_change_factor

    def retrieve(self, measurement_l1, forward_model: ForwardModel, retrieval_target: RetrievalTarget):
        output_dict = dict()

        y_meas_dict = retrieval_target.measurement_vector(measurement_l1)

        y_meas = y_meas_dict['y']
        good_meas = ~np.isnan(y_meas)
        y_meas = y_meas[good_meas]

        if 'y_error' in y_meas_dict:
            # Have measurement error
            if len(np.shape(y_meas_dict['y_error'])) == 1:
                # Only supplied is the diagonal of the error elements
                Sy = sparse.csc_matrix(sparse.diags(y_meas_dict['y_error'][good_meas], 0))
                inv_Sy = sparse.csc_matrix(sparse.diags(1/y_meas_dict['y_error'][good_meas], 0))
            else:
                Sy = y_meas_dict['y_error']
                inv_Sy = np.linalg.inv(Sy)
        else:
            # No user supplied error, use identity matrix
            Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))
            inv_Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))

        x_a = retrieval_target.apriori_state()
        initial_guess = retrieval_target.state_vector()

        inv_Sa = retrieval_target.inverse_apriori_covariance()

        if inv_Sa is None:
            # No apriori covariance/regularization
            # Use initial guess to make the matrices as x_a might be None as well
            inv_Sa = np.zeros((len(initial_guess), len(initial_guess)))

        if x_a is None:
            x_a = np.zeros_like(initial_guess)

        xs = []
        xs.append(initial_guess)
        ys = []
        chi_sq_prev = None
        for iter_idx in range(self._max_iter):
            x = retrieval_target.state_vector()
            forward_l1 = forward_model.calculate_radiance()

            y_ret_dict = retrieval_target.measurement_vector(forward_l1)

            K = y_ret_dict['jacobian'][good_meas, :]
            y_ret = y_ret_dict['y'][good_meas]

            ys.append(y_ret)

            # Left side of rodgers equation
            A = K.T @ inv_Sy @ K + self._lm_damping * np.diag(np.diag(K.T @ K)) + inv_Sa
            A_without_lm = K.T @ inv_Sy @ K + inv_Sa

            # Right side of rodgers equation
            B = K.T @ inv_Sy @ (y_meas - y_ret) - inv_Sa @ (x - x_a)

            if sparse.issparse(A):
                delta_x = np.linalg.solve(A.toarray(), B)
            else:
                delta_x = np.linalg.solve(A, B)

            if sparse.issparse(A_without_lm):
                delta_x_without_lm = np.linalg.solve(A_without_lm.toarray(), B)
            else:
                delta_x_without_lm = np.linalg.solve(A_without_lm, B)

            x_new = x + delta_x

            chi_sq_only_meas = (y_meas - y_ret).T @ inv_Sy @ (y_meas - y_ret)
            chi_sq = chi_sq_only_meas + (x_a - x).T @ inv_Sa @ (x_a - x)

            chi_sq_only_meas_linear = (y_meas - y_ret - K @ delta_x_without_lm).T @ inv_Sy @\
                                      (y_meas - y_ret - K @ delta_x_without_lm)

            chi_sq_linear = (y_meas - y_ret - K @ delta_x_without_lm).T @ inv_Sy @\
                            (y_meas - y_ret - K @ delta_x_without_lm) +\
                            (x_a - x - delta_x_without_lm).T @ inv_Sa @ (x_a - x - delta_x_without_lm)

            chi_sq_only_meas /= len(y_meas)
            chi_sq /= len(y_meas)

            chi_sq_only_meas_linear /= len(y_meas)
            chi_sq_linear /= len(y_meas)

            logging.info('chi_sq:' + str(chi_sq) + '(' + str(chi_sq_only_meas) + ')')
            logging.info('expected_chi_sq:' + str(chi_sq_linear) + '(' + str(chi_sq_only_meas_linear) + ')')

            retrieval_target.update_state(x_new)

            if chi_sq_prev is not None and chi_sq_prev < chi_sq:
                # Iteration was worse
                if self._iterative_update_lm:
                    self._lm_damping *= self._lm_change_factor
                if self._retreat_lm:
                    retrieval_target.update_state(prev_state)
                logging.info("Iteration was worse increasing LM factor to {}".format(self._lm_damping))
            elif chi_sq_prev is not None and chi_sq < chi_sq_prev:
                self._lm_damping /= self._lm_change_factor
                logging.info("Iteration was better decreasing LM factor to {}".format(self._lm_damping))

            xs.append(retrieval_target.state_vector())
            # Store previous values in case we need to retreat
            chi_sq_prev = chi_sq
            prev_state = copy(retrieval_target.state_vector())

        if self._max_iter > 0:
            # Calculate the solution covariance and averaging kernels
            try:
                S = np.linalg.inv(A_without_lm)
            except np.linalg.LinAlgError:
                S = np.linalg.pinv(A_without_lm)

            G = S @ K.T @ inv_Sy
            A = G @ K
            meas_error_covar = G @ (Sy.dot(G.T))
        else:
            S = np.zeros((len(x_a), len(x_a)))
            G = None
            A = np.zeros((len(x_a), len(x_a)))
            meas_error_covar = np.zeros((len(x_a), len(x_a)))

        output_dict['xs'] = xs
        output_dict['gain_matrix'] = G
        output_dict['averaging_kernel'] = A

        output_dict['error_covariance_from_noise'] = meas_error_covar
        output_dict['solution_covariance'] = S

        output_dict['ys'] = ys
        output_dict['y_meas'] = y_meas

        if self._max_iter > 0:
            output_dict['chi_sq_meas'] = chi_sq_only_meas
            output_dict['chi_sq_meas_linear'] = chi_sq_only_meas_linear

            output_dict['chi_sq'] = chi_sq
            output_dict['chi_sq_linear'] = chi_sq_linear
        else:
            output_dict['chi_sq_meas'] = None
            output_dict['chi_sq_meas_linear'] = None

            output_dict['chi_sq'] = None
            output_dict['chi_sq_linear'] = None

        return output_dict
